"use strict";
var __read = (this && this.__read) || function (o, n) {
    var m = typeof Symbol === "function" && o[Symbol.iterator];
    if (!m) return o;
    var i = m.call(o), r, ar = [], e;
    try {
        while ((n === void 0 || n-- > 0) && !(r = i.next()).done) ar.push(r.value);
    }
    catch (error) { e = { error: error }; }
    finally {
        try {
            if (r && !r.done && (m = i["return"])) m.call(i);
        }
        finally { if (e) throw e.error; }
    }
    return ar;
};
var __spreadArray = (this && this.__spreadArray) || function (to, from, pack) {
    if (pack || arguments.length === 2) for (var i = 0, l = from.length, ar; i < l; i++) {
        if (ar || !(i in from)) {
            if (!ar) ar = Array.prototype.slice.call(from, 0, i);
            ar[i] = from[i];
        }
    }
    return to.concat(ar || Array.prototype.slice.call(from));
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.findNameBasedMatch = exports.variationsOf = exports.fixDashedLicenseInfo = exports.isKnownExceptionId = exports.isKnownLicenseId = exports.correctExceptionId = exports.correctLicenseId = exports.exceptions = exports.licenses = void 0;
var spdx_correct_1 = __importDefault(require("spdx-correct"));
var permutations_1 = require("../utils/permutations");
var data_1 = require("./data");
Object.defineProperty(exports, "licenses", { enumerable: true, get: function () { return data_1.licenses; } });
Object.defineProperty(exports, "exceptions", { enumerable: true, get: function () { return data_1.exceptions; } });
var idEndsWithVersionNumber = function (id) {
    return !!id.match(/^.*\d+(\.\d+)*$/);
};
var removeVersionNumberFromId = function (id) {
    return idEndsWithVersionNumber(id) ? id.replace(/(v|\-)?\d+(\.\d+)*$/, '') : id;
};
var countVersionsOf = function (ids, id) {
    var prefix = removeVersionNumberFromId(id);
    return ids.filter(function (candidateId) { return candidateId.startsWith(prefix); }).length;
};
var createMapOfIds = function (ids) {
    var mutilateId = function (id) { return id.toUpperCase().replace(/[^0-9a-zA-Z\+]/g, ''); };
    var listOfOfficialIds = ids;
    var mapOfLowercaseIds = new Map();
    var mapOfMutilatedIds = new Map();
    var mapOfSingleVersionIdsWithoutExplicitVersion = new Map();
    listOfOfficialIds.forEach(function (id) {
        mapOfLowercaseIds.set(id.toLowerCase(), id);
        mapOfMutilatedIds.set(mutilateId(id), id);
        if (idEndsWithVersionNumber(id) && countVersionsOf(listOfOfficialIds, id) === 1) {
            mapOfSingleVersionIdsWithoutExplicitVersion.set(removeVersionNumberFromId(id.toLowerCase()), id);
        }
    });
    var getExactMatch = function (id) { return listOfOfficialIds.find(function (x) { return x === id; }); };
    var getCaseInsensitiveMatch = function (id) {
        var lowercaseId = id.toLowerCase();
        var variationsToTest = [
            lowercaseId,
            lowercaseId.replace(/\s+/g, '-').replace(/(\d+)$/, '$1.0')
        ];
        return variationsToTest.map(function (x) { return mapOfLowercaseIds.get(x); }).filter(function (m) { return !!m; })[0];
    };
    var getFuzzyMatch = function (id) { return mapOfMutilatedIds.get(mutilateId(id)); };
    var getSingleVersionMatch = function (id) { return mapOfSingleVersionIdsWithoutExplicitVersion.get(id); };
    var get = function (id) {
        return getExactMatch(id) || getCaseInsensitiveMatch(id) || getFuzzyMatch(id) || getSingleVersionMatch(id);
    };
    return { get: get };
};
var mapOfKnownLicenses = createMapOfIds(data_1.licenses.licenses.map(function (lic) { return lic.licenseId; }));
var mapOfKnownExceptions = createMapOfIds(data_1.exceptions.exceptions.map(function (e) { return e.licenseExceptionId; }));
var aliasesForLicenseIds = (function () {
    var mapOfAliases = new Map();
    // Aliases for the BSD-2-Clause license:
    mapOfAliases.set('freebsd', 'BSD-2-Clause');
    mapOfAliases.set('freebsd license', 'BSD-2-Clause');
    mapOfAliases.set('freebsd license', 'BSD-2-Clause');
    mapOfAliases.set('simplified bsd license', 'BSD-2-Clause');
    // Aliases for the BSD-3-Clause license:
    mapOfAliases.set('new bsd license', 'BSD-3-Clause');
    mapOfAliases.set('modified bsd license', 'BSD-3-Clause');
    // Aliases for the 0BSD license:
    mapOfAliases.set('zero-clause bsd', '0BSD');
    mapOfAliases.set('free public license', '0BSD');
    mapOfAliases.set('free public license 1.0.0', '0BSD');
    mapOfAliases.set('free public license 1.0', '0BSD');
    mapOfAliases.set('bsd0', '0BSD');
    // Aliases for the Apache-1.1 license:
    mapOfAliases.set('apache1.1', 'Apache-1.1');
    mapOfAliases.set('apache-1.1', 'Apache-1.1');
    mapOfAliases.set('apache software license', 'Apache-1.1');
    mapOfAliases.set('apache software license 1.1', 'Apache-1.1');
    mapOfAliases.set('apache software license version 1.1', 'Apache-1.1');
    // GNU Free Documentation License v1.1
    //mapOfAliases.set('GFDL-1.1', 'GNU Free Documentation License v1.1')
    return mapOfAliases;
})();
var aliasesForExceptions = (function () {
    var mapOfAliases = new Map();
    mapOfAliases.set('qwt license 1.0', 'Qwt-exception-1.0');
    mapOfAliases.set('cpe', 'Classpath-exception-2.0');
    return mapOfAliases;
})();
var mapLicenseAlias = function (alias) {
    return aliasesForLicenseIds.get(alias.toLowerCase());
};
var mapLicenseId = function (id) {
    return mapOfKnownLicenses.get(id.toLowerCase());
};
var shouldIgnoreCorrection = function (input, output) {
    // known bug in spdx-correcting e.g. "GNU Free Documentation License v1.1" into GPL-3.0-or-later:
    return (output === null || output === void 0 ? void 0 : output.startsWith('GPL-')) && input.includes('Free Documentation License');
};
var fixLicenseId = function (id, upgradeGPLVariants) {
    // Don't try to correct license identifiers that are syntactically legit on the
    // surface and have an explicit version number such as "3.0" or "2.1":
    if (id.match(/(\w+(\-\w+)*)\-\d+(\.\d+)+(\-\w+)*/)) {
        return id;
    }
    // Use `spdx-correct` to try and correct identifiers that we haven't fixed already:
    var corrected = (0, spdx_correct_1.default)(id, { upgrade: upgradeGPLVariants });
    if (corrected && !shouldIgnoreCorrection(id, corrected)) {
        return corrected;
    }
    return null;
};
var mapExceptionAlias = function (alias) {
    return aliasesForExceptions.get(alias.toLowerCase());
};
var mapExceptionId = function (id) {
    return mapOfKnownExceptions.get(id.toLowerCase());
};
var fixExceptionid = function (id) {
    var mutations = [
        function (id) { return id.replace(/\s+/, '-'); },
        function (id) { return id.replace(/\s+version\s+/i, ' '); },
        function (id) { return id.replace(/([^.])(\d+)$/, '$1$2.0'); },
        function (id) { return id.replace(/^GNU (.*)$/i, '$1'); },
        function (id) { return id.replace(/(.+)\s+\(.+?\)((v|version )?\d\.\d)?/gi, '$1$2'); },
    ];
    var permutations = (0, permutations_1.permutationsOf)(mutations, 3);
    var potentialIdentifiers = permutations.map(function (combo) {
        var initialValue = combo[0](id);
        return combo.slice(1).reduce(function (prev, curr) { return curr(prev); }, initialValue);
    });
    var matchedIds = __spreadArray([], __read(new Set(potentialIdentifiers.map(mapExceptionId).filter(function (id) { return !!id; }))), false);
    if (matchedIds.length > 1)
        console.warn("fixExceptionId(".concat(JSON.stringify(id), ") found ").concat(matchedIds.length, " matches through mutations: ").concat(JSON.stringify(matchedIds)));
    return matchedIds[0];
};
var mapExceptionIdByAssociatedLicense = function (id, associatedLicense) {
    var expandLicenses = function (identifiers) {
        var expandedList = __spreadArray([], __read(identifiers), false);
        identifiers.forEach(function (id) {
            if (id.match(/[^\+]\+$/)) {
                expandedList.push(id.replace(/\+$/, '-or-later'));
                expandedList.push(id.replace(/\+$/, ''));
            }
            else if (id.match(/-or-later$/)) {
                expandedList.push(id.replace(/-or-later$/, ''));
            }
            else if (id.match(/-only$/)) {
                expandedList.push(id.replace(/-only$/, ''));
            }
            else if (id.match(/[AL]?GPL\-\d\.\d$/)) {
                expandedList.push("".concat(id, "-only"));
                expandedList.push("".concat(id, "-or-later"));
            }
        });
        return expandedList;
    };
    if (associatedLicense) {
        var possibleMatches = data_1.exceptions.exceptions.filter(function (e) {
            var hasRelation = function () {
                var relatedLicenses = expandLicenses(e.relatedLicenses);
                return relatedLicenses.length === 0 || relatedLicenses.includes(associatedLicense);
            };
            var looksSimilar = function () {
                return e.licenseExceptionId.toLowerCase().startsWith(id.toLowerCase());
            };
            return looksSimilar() && hasRelation();
        }).map(function (e) { return e.licenseExceptionId; }).sort();
        if (possibleMatches.length > 0) {
            return possibleMatches[0];
        }
    }
    return undefined;
};
/**
 * Fix the given license identifier, if possible, or return it unchanged.
 *
 * @param identifier SPDX license identifier to "fix" if possible
 * @returns A corrected SPDX license identifier or the original, if no suitable "fix" was found
 */
function correctLicenseId(identifier, upgradeGPLVariants) {
    var mapLicense = function (id) { return mapLicenseId(id) || mapLicenseAlias(id); };
    // The `spdx-correct` package does not have fixes for everything (e.g. a common misspelling for the
    // zero-clause variant of the BSD license or for names such as "The Simplified BSD license") so we'll
    // check for a list of known aliases ourselves:
    var applyAliases = function (id) {
        var lowercaseId = id.replace(/^the /i, '');
        return mapLicense(lowercaseId) || mapLicense("the ".concat(lowercaseId)) || id;
    };
    // The `spdx-correct` package will coerce "GPL-3.0" into "GPL-3.0-or-later", although
    // "GPL-3.0-only" would be more true to the original intent:
    var applyGPLFixes = function (id) {
        var onlySuffix = upgradeGPLVariants ? '-only' : '';
        if (upgradeGPLVariants && mapLicense(id))
            return mapLicense("".concat(id).concat(onlySuffix)) || id;
        if (id.toUpperCase() === 'GPL')
            return mapLicense("GPL-3.0".concat(onlySuffix)) || id;
        if (id.toUpperCase() === 'LGPL')
            return mapLicense("LGPL-3.0".concat(onlySuffix)) || id;
        if (id.toUpperCase() === 'AGPL')
            return mapLicense("AGPL-3.0".concat(onlySuffix)) || id;
        // Expand a single-digit "GPL2+" or "GPLv3" into a "-2.0+" or "-3.0" with the trailing zero
        if (id.match(/^([AL]?GPL)([\-vV]?)(\d)(\+?)$/)) {
            return applyGPLFixes(id.replace(/^([AL]?GPL)([\-vV]?)(\d)(\+?)$/, '$1-$3.0$4'));
        }
        // Correct the '-and-later' to the canonical '-or-later'
        if (id.match(/^([AL]?GPL.*)(\-and\-later)$/)) {
            var candidateId = id.replace(/([AL]?GPL.*)(\-and\-later)$/, '$1-or-later');
            return applyGPLFixes(candidateId) || id;
        }
        // Expand the '+' syntax to the canonical '-or-later'
        if (id.match(/^([AL]?GPL[\-vV]?\d)(\+)$/)) {
            var candidateId = id.replace(/([AL]?GPL[\-vV]?\d)(\+)$/, '$1.0-or-later');
            return mapLicense(candidateId) || id;
        }
        return id;
    };
    // Expand the "+" syntax to "-or-later" if one exists:
    var expandPlus = function (id) {
        if (id.match(/[^\+]\+$/)) {
            var orLaterId = id.replace(/\+$/, '-or-later');
            return mapLicense(orLaterId) || id;
        }
        return id;
    };
    // If the license identifier contains a "-with-" substring, it's an expression, not
    // a single identifier so let's not try to "fix" it at this point:
    if (identifier.includes('-with-'))
        return identifier;
    var aliased = applyAliases(identifier);
    if (aliased !== identifier) {
        // the identifier was successfully aliased – let's retry with the alias!
        return correctLicenseId(aliased, upgradeGPLVariants);
    }
    var id = expandPlus(applyGPLFixes(aliased));
    return mapLicense(id) || fixLicenseId(id, upgradeGPLVariants) || id;
}
exports.correctLicenseId = correctLicenseId;
/**
 * Fix the given exception identifier, if possible, or return it unchanged.
 *
 * @param identifier SPDX exception identifier to "fix" if possible
 * @returns A corrected SPDX exception identifier or the original, if no suitable "fix" was found
 */
function correctExceptionId(identifier, associatedLicense) {
    var removeExtras = function (id) {
        return id.replace(/^the\s+/i, '').replace(/\s+/, ' ');
    };
    var applyAliases = function (id) {
        var lowercaseId = id.toLowerCase();
        var lowercaseIdWithoutWordVersion = lowercaseId.replace(/\s+version\s+(\d+)/, ' $1');
        return mapExceptionAlias(lowercaseId) || mapExceptionAlias(lowercaseIdWithoutWordVersion) || id;
    };
    var id = applyAliases(removeExtras(identifier));
    return mapExceptionId(id) || fixExceptionid(id) || mapExceptionIdByAssociatedLicense(id, associatedLicense) || id;
}
exports.correctExceptionId = correctExceptionId;
function isKnownLicenseId(id) {
    return !!mapLicenseId(id);
}
exports.isKnownLicenseId = isKnownLicenseId;
function isKnownExceptionId(id) {
    return !!mapExceptionId(id);
}
exports.isKnownExceptionId = isKnownExceptionId;
function fixDashedLicenseInfo(node, upgradeGPLVariants) {
    if (!node.exception && node.license.includes('-with-') && !isKnownLicenseId(node.license)) {
        // let's see if we can map the parts to actual known identifiers
        var _a = __read(node.license.split('-with-'), 2), splitLicense = _a[0], splitException = _a[1];
        var fixedLicense = correctLicenseId(splitLicense, upgradeGPLVariants);
        var fixedException = correctExceptionId(splitException, fixedLicense);
        if (isKnownLicenseId(fixedLicense)) {
            return { license: fixedLicense, exception: fixedException };
        }
    }
    return node;
}
exports.fixDashedLicenseInfo = fixDashedLicenseInfo;
var variationsOf = function (name, upgradeGPLVariants) {
    var versionPrefixVariationsOf = function (_value) { return [' version ', ' v']; };
    var namePrefixVariationsOf = function (value) {
        var namePrefixVariations = [value];
        if (value.startsWith('gnu ')) {
            namePrefixVariations.push(value.substring('gnu '.length));
        }
        else if (value.match(/^((lesser|affero) )?general public license/)) {
            namePrefixVariations.push("gnu ".concat(value));
        }
        return namePrefixVariations;
    };
    var suffixVariationsOf = function (value) {
        var fixedAlternatives = [' or greater', ' or later', ' or newer'];
        if (value === undefined) {
            if (upgradeGPLVariants) {
                return ['', ' only'];
            }
            else {
                return [''];
            }
        }
        else if (fixedAlternatives.includes(value)) {
            return __spreadArray([value], __read(fixedAlternatives.filter(function (x) { return x !== value; })), false);
        }
        else {
            return [value];
        }
    };
    var versionNumberVariationsOf = function (value) {
        var versionNumberVariations = [value];
        if (value.endsWith('.0')) {
            // For "3.1.0", let's also try "3.1"
            versionNumberVariations.push(value.replace(/\.0$/, ''));
            if (value.endsWith('.0.0')) {
                // For "3.0.0", let's also try "3"
                versionNumberVariations.push(value.replace(/\.0\.0$/, ''));
            }
        }
        else if (!value.includes('.')) {
            // For "3", let's also try "3.0"
            versionNumberVariations.push("".concat(value, ".0"));
        }
        return versionNumberVariations;
    };
    var variations = [name];
    var versionMatch = name.match(/^(.+)(\sversion\s|\sv)(\d+(\.\d+)*)(\+|(\s.*))?$/);
    if (versionMatch) {
        var namePrefixVariations = namePrefixVariationsOf(versionMatch[1]);
        var suffixVariations_1 = suffixVariationsOf(versionMatch[5]);
        var versionPrefixVariations_1 = versionPrefixVariationsOf(versionMatch[2]);
        var versionNumberVariations_1 = versionNumberVariationsOf(versionMatch[3]);
        namePrefixVariations.forEach(function (namePrefix) {
            versionPrefixVariations_1.forEach(function (versionPrefix) {
                versionNumberVariations_1.forEach(function (versionNumber) {
                    suffixVariations_1.forEach(function (suffix) {
                        variations.push("".concat(namePrefix).concat(versionPrefix).concat(versionNumber).concat(suffix));
                    });
                });
            });
        });
    }
    return variations.sort(function (a, b) {
        var lengthBasedSort = b.length - a.length;
        if (lengthBasedSort !== 0) {
            // string length is the primary sort key
            return lengthBasedSort;
        }
        else {
            // between strings of equal length, sort alphabetically
            // (with uppercase "A" coming before lowercase "a")
            return a.localeCompare(b, undefined, { caseFirst: 'upper' });
        }
    });
};
exports.variationsOf = variationsOf;
/**
 * Try to identify the license by comparing the given text with the official names of
 * known licenses in the SPDX data.
 *
 * @param text The text that could be a license name
 * @returns The {@link License}, or `undefined`
 */
var findNameBasedMatch = function (text, upgradeGPLVariants) {
    var normalizeName = function (name) {
        return name.toLowerCase().replace(/,/g, ' ').replace(/\s+/, ' ');
    };
    var lowercaseText = normalizeName(text);
    var variations = (0, exports.variationsOf)(lowercaseText, upgradeGPLVariants);
    return data_1.licenses.licenses.find(function (x) { return variations.includes(normalizeName(x.name)); });
};
exports.findNameBasedMatch = findNameBasedMatch;
