"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.parse = exports.LiberalParser = void 0;
var assert_1 = __importDefault(require("assert"));
var LiberalParser = __importStar(require("../codegen/parser_liberal"));
exports.LiberalParser = LiberalParser;
var licenses_1 = require("../licenses");
var extractErrorMessage = function (tree) {
    if (tree.errs.length === 0)
        return undefined;
    return tree.errs.map(function (element) { return element.toString(); }).join('\n');
};
function parse(input, upgradeGPLVariants) {
    var _a;
    var convertAnOrLaterOrLaterSpecialCase = function (node) {
        var isLicenseInfo = function (node) {
            return node.license;
        };
        var isALater = function (node) {
            return ['later', 'newer', 'greater'].includes(node.license);
        };
        var isADashOrDashLaterLicense = function (node) {
            return node.license.endsWith('-or-later');
        };
        if (isLicenseInfo(node.left) && isLicenseInfo(node.right)) {
            var left = node.left;
            var right = node.right;
            if (isALater(right) && !isADashOrDashLaterLicense(left)) {
                var dashedVersion = "".concat(left.license, "-or-later");
                if ((0, licenses_1.isKnownLicenseId)(dashedVersion))
                    return __assign(__assign({}, left), { license: dashedVersion });
                else
                    return __assign(__assign({}, left), { license: "".concat(left.license, " or later") });
            }
            else if (isALater(left) && !isADashOrDashLaterLicense(right)) {
                var dashedVersion = "".concat(right.license, "-or-later");
                if ((0, licenses_1.isKnownLicenseId)(dashedVersion))
                    return __assign(__assign({}, right), { license: dashedVersion });
                return __assign(__assign({}, right), { license: "".concat(right.license, " or later") });
            }
            else if (isALater(right) && isADashOrDashLaterLicense(left)) {
                return left;
            }
            else if (isALater(left) && isADashOrDashLaterLicense(right)) {
                return right;
            }
        }
        return node;
    };
    var reduceNode = function (node) {
        if (!node) {
            return undefined;
        }
        else if (node.kind === LiberalParser.ASTKinds.or_expression) {
            var conjunction = {
                conjunction: 'or',
                left: reduceNode(node.left),
                right: reduceNode(node.right)
            };
            return convertAnOrLaterOrLaterSpecialCase(conjunction);
        }
        else if (node.kind === LiberalParser.ASTKinds.and_expression) {
            return {
                conjunction: 'and',
                left: reduceNode(node.left),
                right: reduceNode(node.right)
            };
        }
        else if (node.kind === LiberalParser.ASTKinds.license_exception) {
            var value = reduceNode(node.value);
            return (0, licenses_1.correctExceptionId)(value);
        }
        else if (node.kind === LiberalParser.ASTKinds.simple_expression) {
            if (node.exception) {
                var license = (0, licenses_1.correctLicenseId)(reduceNode(node.value), upgradeGPLVariants);
                var exception = (0, licenses_1.correctExceptionId)(reduceNode(node.exception), license);
                return { license: license, exception: exception };
            }
            else {
                var license = (0, licenses_1.correctLicenseId)(reduceNode(node.value), upgradeGPLVariants);
                if (!(0, licenses_1.isKnownLicenseId)(license) && license.includes('-with-')) {
                    return (0, licenses_1.fixDashedLicenseInfo)({ license: license }, upgradeGPLVariants);
                }
                return { license: license };
            }
        }
        else if (node.kind === LiberalParser.ASTKinds.words_expression) {
            var head = node.value.prefix.head.value;
            var tail = node.value.prefix.tail ? reduceNode(node.value.prefix.tail.rest) : '';
            return (head + ' ' + tail).trim();
        }
        else if (node.kind === LiberalParser.ASTKinds.words) {
            if (node.tail) {
                return (node.head.value + ' ' + reduceNode(node.tail.rest)).trim();
            }
            else {
                return node.head.value;
            }
        }
        else if (node.kind === LiberalParser.ASTKinds.wrapped_expression) {
            return reduceNode(node.value);
        }
        /* istanbul ignore next */ assert_1.default.fail("convertNodeWithCorrections() did not recognize input: ".concat(JSON.stringify(node)));
    };
    var p = new LiberalParser.Parser(input);
    var tree = p.parse();
    var expression = reduceNode((_a = tree.ast) === null || _a === void 0 ? void 0 : _a.value);
    return { parse: tree, expression: expression, error: extractErrorMessage(tree) };
}
exports.parse = parse;
